#!/usr/bin/env python

from collections import defaultdict

from distributions import *
from schedcat.model.LLCLayout import *
from schedcat.model.LLC_params import *
import schedcat.model.ohc as ohc

from pulp import *
import gurobipy

import numpy as np
import time

def LPLLCsets(ts_mcx, vartype, csch, olap = True):
    ts_mcx.setScheme(csch)
#     sched = ts_mcx.get_sched_app()

#     CORES = ts_mcx.m
    
    capped_wss = {  'B' : [512 for i in range(CORES)],
                    'C' : 512}
                    
    sched = ts_mcx.get_sched_all()
    ts_mcx.A = ts_mcx.level('A')
    ts_mcx.B = ts_mcx.level('B')
    ts_mcx.C = ts_mcx.level('C')
    
    UA_Adata = np.zeros((CORES, csch.mS+1))
    UA_Bdata = np.zeros((CORES, csch.mS+1))
    UA_Cdata = np.zeros((CORES, csch.mS+1))
    
    UB_Bdata = np.zeros((CORES, csch.mS+1))
    UB_Cdata = np.zeros((CORES, csch.mS+1))
    
    minAperiods = []
    for core in range(0, CORES):
        if len(ts_mcx.A.core(core)) == 0:
            minAperiods.append(0)
        else:
            minAperiods.append(min([t.period for t in ts_mcx.A.core(core)]))
    
    for s in range(0,MAX_SETS+1):
    	wsch = scheme([area(csch.A[i].W, s) for i in range(0, CORES)],
                      [area(csch.B[i].W, s) for i in range(0, CORES)],
                      csch.C, csch.mW, csch.mS)

        ts_mcx.setScheme(wsch)
        for core in range(0, CORES):                      
            UA_Adata[core][s] = ts_mcx.A.core(core).utilization(atlevel = 'A', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
            UA_Bdata[core][s] = ts_mcx.A.core(core).utilization(atlevel = 'B', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
            UA_Cdata[core][s] = ts_mcx.A.core(core).utilization(atlevel = 'C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
            
            UB_Bdata[core][s] = ts_mcx.B.core(core).utilization(atlevel = 'B', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
            UB_Cdata[core][s] = ts_mcx.B.core(core).utilization(atlevel = 'C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
                
        
        UAi = defaultdict()
        UAi['B'] = [UA_Bdata[core][s] for core in range(0, CORES)]
        UAi['C'] = [UA_Cdata[core][s] for core in range(0, CORES)]
        
        UAi = ts_mcx.secondaryInflateConst(UAi, minAperiods = minAperiods)
        for core in range(0, CORES):
            UA_Bdata[core][s] = UAi['B'][core]
            UA_Cdata[core][s] = UAi['C'][core]
        
    UC_C = ts_mcx.C.utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
    h = ts_mcx.C.max_utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
    H = ts_mcx.C.utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss, heaviest = CORES-1)
    WO = [max(0, csch.A[i].W+csch.B[i].W+csch.C.W-csch.mW) for i in range(0, CORES)]
    
    problem = LpProblem("Hybrid",LpMinimize)
    
    UA_A = LpVariable.matrix("UA_A", range(CORES), lowBound = 0)
    UA_B = LpVariable.matrix("UA_B", range(CORES), lowBound = 0)
    UA_C = LpVariable.matrix("UA_C", range(CORES), lowBound = 0)
    
    UB_B = LpVariable.matrix("UB_B", range(CORES), lowBound = 0)
    UB_C = LpVariable.matrix("UB_C", range(CORES), lowBound = 0)
    
    S = LpVariable.matrix("S", range(CORES), lowBound = 0, cat=vartype)
    
    for Si in S:
        Si.upBound = MAX_SETS
    
    problem += lpSum(S) <= MAX_SETS
            
    for i in range(0, CORES):
        UA_A[i].upBound = 1
        UA_B[i].upBound = 1
        UA_C[i].upBound = 1
        
        UB_B[i].upBound = 1
        UB_C[i].upBound = 1
        
        if not len(ts_mcx.A.core(i)) is 0:
            invminp = 1.0/min([t.period for t in ts_mcx.A.core(i)])
        else:
            invminp = 0
            
        tckinfw = ohc.PTCK/(ohc.PTCK-ohc.ohcvals['Max']['TCK'])
        tckinfa = ohc.PTCK/(ohc.PTCK-ohc.ohcvals['Avg']['TCK'])
        IA_B = ohc.CPMD*csch.B[i]*S[i]*invminp*tckinfw if olap else 0
        IA_C = ohc.CPMD*WO[i]*csch.S[i]*invminp*tckinfa if olap else 0
        
        for j in range(0,MAX_SETS):
            C1 = (UA_Adata[i][j+1]-UA_Adata[i][j])
            C2 = UA_Adata[i][j]-j*C1
            problem += (UA_A[i]-C1*S[i] >= C2)
            C1 = (UA_Bdata[i][j+1]-UA_Bdata[i][j])
            C2 = UA_Bdata[i][j]-j*C1
            problem += (UA_B[i]-C1*S[i] >= C2 + IA_B)
            C1 = (UA_Cdata[i][j+1]-UA_Cdata[i][j])
            C2 = UA_Cdata[i][j]-j*C1
            problem += (UA_C[i]-C1*S[i] >= C2 + IA_C)
            
            C1 = (UB_Bdata[i][j+1]-UB_Bdata[i][j])
            C2 = UB_Bdata[i][j]-j*C1
            problem += (UB_B[i]-C1*S[i] >= C2)
            C1 = (UB_Cdata[i][j+1]-UB_Cdata[i][j])
            C2 = UB_Cdata[i][j]-j*C1
            problem += (UB_C[i]-C1*S[i] >= C2)
    
    
    objset = False
    for c in range(0,CORES):
        if sched['A'][c]:
            problem += UA_A[c] <= 1
        elif not objset:
            problem += UA_A[c]
            objset = True
        if sched['B'][c]:
            problem += UA_B[c] + UB_B[c] <= 1
        elif not objset:
            problem += UA_B[c] + UB_B[c]
            objset = True
            
    if sched['C']:
        problem += lpSum(UA_C) + lpSum(UB_C) + UC_C <= CORES
    elif not objset:
        problem += lpSum(UA_C) + lpSum(UB_C) + UC_C
        objset = True
        
    eps = 1e-6
    
    if sched['H']:
        problem += lpSum(UA_C) + lpSum(UB_C) + (CORES-1)*h + H <= CORES-eps
    elif not objset:
        problem += lpSum(UA_C) + lpSum(UB_C) + (CORES-1)*h + H
    else:
        problem += lpSum(UA_C) + lpSum(UB_C) + UC_C
    
    # print problem 
    start = time.clock() 
    status = problem.solve(GUROBI(msg=0, timeLimit = 480))
    fin = time.clock()
    
    rntm = -1
    if LpStatus[status] != 'Not Solved':
        rntm = fin-start

    if LpStatus[status] == 'Optimal':
        ps = scheme([area(csch.A[i].W, S[i].varValue) for i in range(0,CORES)],
                    [area(csch.B[i].W, S[i].varValue) for i in range(0,CORES)],
                    csch.C, MAX_WAYS, MAX_SETS)
        
        ps.A = [val.rounded() for val in ps.A]
        ps.B = [val.rounded() for val in ps.B]
        ps.C = ps.C.rounded()
        
        return ps, rntm
    if LpStatus[status] == 'Not Solved':
        return "TimeOut", rntm
        
    return None, rntm

def LPLLCways(ts_mcx, vartype, csch, olap = True, debug = False):
    CORES = ts_mcx.m
    
    capped_wss = {  'B' : [512 for i in range(CORES)],
                    'C' : 512}
    mWC = csch.mW*ts_mcx.wg
#     ts_mcx.setScheme(csch)
#     sched = ts_mcx.get_sched_app()F
    sched = ts_mcx.get_sched_all()
    ts_mcx.A = ts_mcx.level('A')
    ts_mcx.B = ts_mcx.level('B')
    ts_mcx.C = ts_mcx.level('C')
    
#     print "debug is " + str(debug)
     
    #----------------------- Gather Data -----------------------------#
    UA_Adata = np.zeros((CORES, mWC+1))
    UA_Bdata = np.zeros((CORES, mWC+1))
    UA_Cdata = np.zeros((CORES, mWC+1))
    
    UB_Bdata = np.zeros((CORES, mWC+1))
    UB_Cdata = np.zeros((CORES, mWC+1))
    UC_Cdata = np.zeros(mWC+1)
    
    hdata = np.zeros(mWC+1)
    Hdata = np.zeros(mWC+1)
    
    minAperiods = []
    for core in range(0, CORES):
        if len(ts_mcx.A.core(core)) == 0:
            minAperiods.append(0)
        else:
            minAperiods.append(min([t.period for t in ts_mcx.A.core(core)]))
    
    for w in range(0,mWC+1):
        wsch = scheme([area(w, csch.A[i].S) for i in range(0, CORES)],
                      [area(w, csch.B[i].S) for i in range(0, CORES)],
                      area(w, csch.C.S), csch.mW, csch.mS)
        ts_mcx.setScheme(wsch)
        for core in range(0, CORES):
            UA_Adata[core][w] = ts_mcx.A.core(core).utilization('A', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
            if debug and core is 0:
                print "UA_Adata on core " + str(core) + " with ways " + str(w) + ": " + str(UA_Adata[core][w])
            UA_Bdata[core][w] = ts_mcx.A.core(core).utilization('B', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
            UA_Cdata[core][w] = ts_mcx.A.core(core).utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
            
            UB_Bdata[core][w] = ts_mcx.B.core(core).utilization('B', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
            UB_Cdata[core][w] = ts_mcx.B.core(core).utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
        
        UAi = defaultdict()
        UAi['B'] = [UA_Bdata[core][w] for core in range(0, CORES)]
        UAi['C'] = [UA_Cdata[core][w] for core in range(0, CORES)]
        
        UAi = ts_mcx.secondaryInflateConst(UAi, minAperiods = minAperiods)
        for core in range(0, CORES):
            UA_Bdata[core][w] = UAi['B'][core]
            UA_Cdata[core][w] = UAi['C'][core]
            if debug and w is 0:
#                 print "UA_Bdata1 on core " + str(core) + " with ways " + str(w) + ": " + str(UA_Bdata[core][w])
                print "UA_Adata on core " + str(core) + " with ways " + str(w) + ": " + str(UA_Adata[core][w])
                print "UA_Bdata on core " + str(core) + " with ways " + str(w) + ": " + str(UA_Bdata[core][w])
                print "UA_Cdata on core " + str(core) + " with ways " + str(w) + ": " + str(UA_Cdata[core][w])
                print "UB_Bdata on core " + str(core) + " with ways " + str(w) + ": " + str(UB_Bdata[core][w])
                print "UB_Cdata on core " + str(core) + " with ways " + str(w) + ": " + str(UB_Cdata[core][w])
        
        UC_Cdata[w] = ts_mcx.C.utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
        
        hdata[w] = ts_mcx.C.max_utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss)
        Hdata[w] = ts_mcx.C.utilization('C', approximate = True, inflated = True, c_taskset = ts_mcx.C, minAperiods = minAperiods, mx_wss = capped_wss, heaviest = CORES-1)
        
        if debug and (w is 16 or w is 0 or w is 1):
            print "C with ways " + str(w) + ": " + str(UC_Cdata[w]) + ", " + str(hdata[w]) + ", " + str(Hdata[w])
    
    #----------------------- Create LP Variables -----------------------------#
    problem = LpProblem("Hybrid",LpMinimize)
    
    UA_A = LpVariable.matrix("UA_A", range(CORES), lowBound = 0)
    UA_B = LpVariable.matrix("UA_B", range(CORES), lowBound = 0)
    UA_C = LpVariable.matrix("UA_C", range(CORES), lowBound = 0)
    
    UB_B = LpVariable.matrix("UB_B", range(CORES), lowBound = 0)
    UB_C = LpVariable.matrix("UB_C", range(CORES), lowBound = 0)
    
    UC_C = LpVariable("UC_C", lowBound = 0)
    
    h = LpVariable("h", lowBound = 0)
    H = LpVariable("H", lowBound = 0)
    
    Wa = LpVariable.matrix("Wa", range(CORES), lowBound = 0, cat=vartype)
    Wb = LpVariable.matrix("Wb", range(CORES), lowBound = 0, cat=vartype)
    Wc = LpVariable("Wc", lowBound = 0, cat=vartype);
    
    WO = LpVariable.matrix("WO", range(CORES), lowBound = 0)
    
    Wc.upBound = mWC
    
    
    #----------------------- Constraint Set 1 -----------------------------#
    for i, c in enumerate(Wa):
        Wa[i].upBound = mWC
        Wb[i].upBound = mWC
        WO[i].upBound = mWC
        
        if olap:
            problem += Wa[i] + Wc <= mWC
            problem += Wb[i] + Wc <= mWC
        else:
            problem += Wa[i] + Wb[i] + Wc <= mWC
            
        problem += WO[i] >= Wa[i] + Wb[i] + Wc - mWC
            
    #----------------------- Constraint Set 2 -----------------------------#
    for i in range(0, CORES):
        UA_A[i].upBound = 1
        UA_B[i].upBound = 1
        UA_C[i].upBound = 1
        
        UB_B[i].upBound = 1
        UB_C[i].upBound = 1
        
        if not len(ts_mcx.A.core(i)) is 0:
            invminp = 1.0/min([t.period for t in ts_mcx.A.core(i)])
        else:
            invminp = 0
            
        tckinfw = ohc.PTCK/(ohc.PTCK-ohc.ohcvals['Max']['TCK'])
        tckinfa = ohc.PTCK/(ohc.PTCK-ohc.ohcvals['Avg']['TCK'])
        IA_B = ohc.CPMD*Wb[i]*csch.B[i].S*invminp*tckinfw if olap else 0
        IA_C = ohc.CPMDavg*WO[i]*csch.B[i].S*invminp*tckinfa if olap else 0
        
        for j in range(0,mWC):
            C1 = (UA_Adata[i][j+1]-UA_Adata[i][j])
            C2 = UA_Adata[i][j]-j*C1
            problem += (UA_A[i]-C1*Wa[i] >= C2)
            C1 = (UA_Bdata[i][j+1]-UA_Bdata[i][j])
            C2 = UA_Bdata[i][j]-j*C1
            problem += (UA_B[i]-C1*Wa[i] >= C2 + IA_B)
            C1 = (UA_Cdata[i][j+1]-UA_Cdata[i][j])
            C2 = UA_Cdata[i][j]-j*C1
            problem += (UA_C[i]-C1*Wa[i] >= C2 + IA_C)
            
            C1 = (UB_Bdata[i][j+1]-UB_Bdata[i][j])
            C2 = UB_Bdata[i][j]-j*C1
            problem += (UB_B[i]-C1*Wb[i] >= C2)
            C1 = (UB_Cdata[i][j+1]-UB_Cdata[i][j])
            C2 = UB_Cdata[i][j]-j*C1
            problem += (UB_C[i]-C1*Wb[i] >= C2)
    
    UC_C.upBound = CORES
    for w in range(0, mWC):
        C1 = (UC_Cdata[w+1]-UC_Cdata[w])
        C2 = UC_Cdata[w]-w*C1
        problem += (UC_C-C1*Wc >= C2)
        
    #----------------------- Constraint Set 3 -----------------------------#
        problem += h >= hdata[w] + (Wc-w)*(hdata[w+1]-hdata[w])
        problem += H >= Hdata[w] + (Wc-w)*(Hdata[w+1]-Hdata[w])
        
    #----------------------- Constraint Set 4 -----------------------------#
    
    objset = False
    for c in range(0,CORES):
        if sched['A'][c]:
            problem += UA_A[c] <= 1
        elif not objset:
            problem += UA_A[c]
            objset = True
        if sched['B'][c]:
            problem += UA_B[c] + UB_B[c] <= 1
        elif not objset:
            problem += UA_B[c] + UB_B[c]
            objset = True
            
    if sched['C']:
        problem += lpSum(UA_C) + lpSum(UB_C) + UC_C <= CORES
    elif not objset:
        problem += lpSum(UA_C) + lpSum(UB_C) + UC_C
        objset = True
        
    eps = 1e-6
    
    if sched['H']:
        problem += lpSum(UA_C) + lpSum(UB_C) + (CORES-1)*h + H <= CORES-eps
    elif not objset:
        problem += lpSum(UA_C) + lpSum(UB_C) + (CORES-1)*h + H
    
    if not objset:
        problem += lpSum(UA_A)
    
    #----------------------- Solve -----------------------------#
    
    start = time.clock() 
    status = problem.solve(GUROBI(msg=0, timeLimit = 480))
    fin = time.clock()
    
    rntm = -1
    
    if LpStatus[status] != 'Not Solved':
        rntm = fin-start
        
    if LpStatus[status] == 'Optimal':
        if debug:
            print "LP results: "
            for i in range(0, CORES):
                print "  Core " + str(i)
                print "    AA: " + str(UA_A[i].varValue)
                print "    AB: " + str(UA_B[i].varValue)
                print "    AC: " + str(UA_C[i].varValue)
            
                print "    BB: " + str(UB_B[i].varValue)
                print "    BC: " + str(UB_C[i].varValue)
            print "  CC: " + str(UC_C.varValue)
            print "  h: " + str(h.varValue)
            print "  H: " + str(H.varValue)
            
            print " C sched: " + str((lpSum(UA_C)+lpSum(UB_C)+UC_C).value())           
            print " H sched AB: " + str((lpSum(UA_C) + lpSum(UB_C)).value()) 
            print " H sched h: " + str(((CORES-1)*h).value())
            print " H sched H: " + str((H).value())
            print " H sched: " + str((lpSum(UA_C) + lpSum(UB_C) + (CORES-1)*h + H).value())
        
        
        ps = scheme([area(Wa[i].varValue, csch.A[i].S) for i in range(0,CORES)],
                    [area(Wb[i].varValue, csch.B[i].S) for i in range(0,CORES)],
                    area(Wc.varValue, MAX_SETS), MAX_WAYS, MAX_SETS)
        
        ps.A = [val.rounded() for val in ps.A]
        ps.B = [val.rounded() for val in ps.B]
        ps.C = ps.C.rounded()
        
        return ps, rntm
    if LpStatus[status] == 'Not Solved':
        return "TimeOut", rntm
    
    return None, rntm