from __future__ import division # use sane division semantics

import copy
import random
import numpy as np

from math   import floor, ceil
from schedcat.util.math    import lcm
from schedcat.util.quantor import forall
from collections import defaultdict

from fractions import Fraction
from schedcat.generator.tasks import uniform
from schedcat.model.LLC_params import *
import matplotlib.pyplot as plt
import matplotlib.axis as axis
import matplotlib.ticker as mtick
from matplotlib import rc


# matplotlib.use('Agg')
# import matplotlib.pyplot as plt
from mpl_toolkits.mplot3d import Axes3D
from LLCLayout import scheme, default_scheme
import ohc

# import mpl_toolkits.axisartist as AA

redfunc = uniform(0, 0.05)

def fluctuate(costs):
    wM = len(costs)
        
    choices = []
    numadj = random.randint(0, 8)
    options = range(1, wM)
    
    for j in range(0, numadj):
        pick = random.randint(0, len(options)-1)
        choices.append(options[pick])
        options.remove(options[pick])
        
    cRange = costs[0]-costs[wM-1]
    fcosts = costs[:]
    for j in options:
        fcosts[j] -= redfunc()*cRange
    
    return fcosts


def mmax(lst):
    if len(lst) == 0:
        return 0
    else:
        return max(lst)

class SporadicTaskMCX(object):
    def __init__(self, exec_cost, period, level, wss, deadline=None, id=None):
        """By default, the construct only creates the bare minimum
        attributes. Other code (or subclasses) can add additional
        attributes (such as response time bounds, resource usage, etc.)
        """
        
        if deadline is None:
            # implicit deadline by default
            deadline = period
        self.period     = period
        self.app_cost   = exec_cost
        self.deadline   = deadline
        self.id         = id
        
        self.level = level
        self.ways = MAX_WAYS if level is 'C' else 0
        self.sets = MAX_WAYS if level is 'C' else 4
       
        self.cost   = exec_cost
        self.wss = wss
        # self.cost = {'C': [[] for i in range(0, len(exec_cost['C']))]}
        # self.cost['C'][self.sets] = fluctuate(self.app_cost['C'][self.sets])
        
        # if not level is 'C':
            # self.cost['B'] = [[] for i in range(0, len(exec_cost['B']))]
            # self.cost['B'][self.sets] = fluctuate(self.app_cost['B'][self.sets])
            # self.cost['B'][self.sets] = [max(self.cost['B'][self.sets][i], self.cost['C'][self.sets][i]) for i in range(0, len(self.cost['B'][self.sets]))]
            
        # if level is 'A':
            # self.cost['A'] = [[] for i in range(0, len(exec_cost['A']))]
            # A_inflation = self.app_cost['A'][self.sets][0]/self.app_cost['B'][self.sets][0]
            # self.cost['A'][self.sets] = [val*A_inflation for val in self.cost['B'][self.sets]]

            
    def implicit_deadline(self):
        return self.deadline == self.period

    def constrained_deadline(self):
        return self.deadline <= self.period
        
    def utilization(self, atlevel, approximate, inflated, minAperiod, c_taskset, mx_wss):
        pet = 0
        if approximate:
            pet += self.app_cost[atlevel][self.sets][self.ways]
        else:
            pet += self.cost[atlevel][self.sets][self.ways]
        
        if inflated:
            if(minAperiod == 0):
                pet = self.inflate(pet, atlevel, 0, c_taskset, mx_wss)
            else:
                pet = self.inflate(pet, atlevel, round(self.period/minAperiod), c_taskset, mx_wss)
        
        return pet/self.period

#     def utilization_q(self):
#         return Fraction(self.cost[self.level][self.sets][self.ways], self.period)
        
    def inflate(self, pet, atlevel, num_slices, c_taskset, mx_wss):
        for ohv in self.ohdict.keys():
            if not (ohv is 'TCK') or (ohv is 'RLL'):
#                 print "adding at (" + self.level + ", " + atlevel + "), " + ohv + ": " + str(self.ohdict[ohv][self.level][atlevel](self, mx_wss, num_slices, c_taskset))
                pet += self.ohdict[ohv][self.level][atlevel](self, mx_wss, num_slices, c_taskset)
        
#         print "c is " + str(len(c_taskset))
        etck = self.ohdict['TCK'][self.level][atlevel](self, mx_wss, num_slices, c_taskset)
        utck = etck/ohc.PTCK
        pet = pet/(1+utck)
        
        pet += self.ohdict['RLL'][self.level][atlevel](self, mx_wss, num_slices, c_taskset)
        
        return pet
    
    ohdict = {
        'CPD' : {'A' : {'A' : lambda self, mx_wss, num_slices, c_taskset : min(self.wss, self.sets*self.ways)*ohc.CPMD   *(num_slices-1)*1.5,
                        'B' : lambda self, mx_wss, num_slices, c_taskset : min(self.wss, self.sets*self.ways)*ohc.CPMD   *(num_slices-1),
                        'C' : lambda self, mx_wss, num_slices, c_taskset : min(self.wss, self.sets*self.ways)*ohc.CPMDavg*(num_slices-1),
                       },
                 'B' : {'B' : lambda self, mx_wss, num_slices, c_taskset : min(mx_wss, self.sets*self.ways)*ohc.CPMD,
                        'C' : lambda self, mx_wss, num_slices, c_taskset : min(mx_wss, self.sets*self.ways)*ohc.CPMDavg,
                       },
                 'C' : {'C' : lambda self, mx_wss, num_slices, c_taskset : min(mx_wss, self.sets*self.ways)*ohc.CPMDavg,
                       },
                },
        'CXS' : {'A' : {'A' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Max']['CXS']*(num_slices)*1.5,
                        'B' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Max']['CXS']*(num_slices),
                        'C' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Avg']['CXS']*(num_slices),
                       },
                 'B' : {'B' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Max']['CXS']*2,
                        'C' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Avg']['CXS']*2,
                       },
                 'C' : {'C' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Avg']['CXS']*2,
                       },
                },
        'SCD' : {'A' : {'A' : lambda self, mx_wss, num_slices, c_taskset : num_slices*ohc.ohcvals['Max']['SCH']*1.5,
                        'B' : lambda self, mx_wss, num_slices, c_taskset : num_slices*ohc.ohcvals['Max']['SCH'],
                        'C' : lambda self, mx_wss, num_slices, c_taskset : num_slices*ohc.ohcvals['Avg']['SCH'],
                       },
                 'B' : {'B' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Max']['SCH']*2,
                        'C' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Avg']['SCH']*2,
                       },
                 'C' : {'C' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Avg']['SCL']*2,
                       },
                },
        'RLQ' : {'A' : {'A' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Max']['RQA']*1.5,
                        'B' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Max']['RQA'],
                        'C' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Avg']['RQA'],
                       },
                 'B' : {'B' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Max']['RQA'],
                        'C' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Avg']['RQA'],
                       },
                 'C' : {'C' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Avg']['RQC'],
                       },
                },
        'RLL' : {'A' : {'A' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Max']['RLA']*1.5,
                        'B' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Max']['RLA'],
                        'C' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Avg']['RLA'],
                       },
                 'B' : {'B' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Max']['RLA'],
                        'C' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Avg']['RLA'],
                       },
                 'C' : {'C' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Avg']['RLC'],
                       },
                },
        'TCK' : {'A' : {'A' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Max']['TCK']*1.5,
                        'B' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Max']['TCK'],
                        'C' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Avg']['TCK'],
                       },
                 'B' : {'B' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Max']['TCK'],
                        'C' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Avg']['TCK'],
                       },
                 'C' : {'C' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Avg']['TCK'],
                       },
                },
        'ISR' : {'A' : {'A' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Max']['ISR']*1.5,
                        'B' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Max']['ISR'],
                        'C' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Avg']['ISR'],
                       },
                 'B' : {'B' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Max']['ISR'],
                        'C' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Avg']['ISR'],
                       },
                 'C' : {'C' : lambda self, mx_wss, num_slices, c_taskset : ohc.ohcvals['Avg']['ISR'],
                       },
                },
        'CSI' : {'A' : {'A' : lambda self, mx_wss, num_slices, c_taskset : sum([ceil(self.period/t.period) for t in c_taskset])*(ohc.ohcvals['Max']['RQC']+ohc.ohcvals['Max']['IPI'])*1.5,
                        'B' : lambda self, mx_wss, num_slices, c_taskset : sum([ceil(self.period/t.period) for t in c_taskset])*(ohc.ohcvals['Max']['RQC']+ohc.ohcvals['Max']['IPI']),
                        'C' : lambda self, mx_wss, num_slices, c_taskset : sum([ceil(self.period/t.period) for t in c_taskset])*(ohc.ohcvals['Avg']['RQC']+ohc.ohcvals['Avg']['IPI']),
                       },
                 'B' : {'B' : lambda self, mx_wss, num_slices, c_taskset : sum([ceil(self.period/t.period) for t in c_taskset])*(ohc.ohcvals['Max']['RQC']+ohc.ohcvals['Max']['IPI']),
                        'C' : lambda self, mx_wss, num_slices, c_taskset : sum([ceil(self.period/t.period) for t in c_taskset])*(ohc.ohcvals['Avg']['RQC']+ohc.ohcvals['Avg']['IPI']),
                       },
                 'C' : {'C' : lambda self, mx_wss, num_slices, c_taskset : 0,
                       },
                },
    }
    
#     def density(self):
#         return self.cost[self.level][self.sets][self.ways] / min(self.period, self.deadline)
# 
#     def density_q(self):
#         return Fraction(self.cost[self.level][self.sets][self.ways], min(self.period, self.deadline))
# 
#     def tardiness(self):
#         """Return this task's tardiness.
#         Note: this function can only be called after some test
#         established a response time bound (response_time must be defined)!
#         """
#         return max(0, self.response_time - self.deadline)
# 
#     def maxjobs(self, interval_length):
#         """Compute the maximum number of jobs that can execute during
#         some interval.
#         Note: this function can only be called after some test
#         established a response time bound (response_time must be defined)!
#         """
#         return int(ceil((interval_length + self.response_time) / self.period))

    def __repr__(self):
        idstr = ", id=%s" % self.id if self.id is not None else ""
        dstr  = ", deadline=%s" % self.deadline if self.deadline != self.period else ""
        return "SporadicTask(%s, %s%s%s)" % (self.cost[self.level][self.sets][self.ways], self.period, dstr, idstr)
        
    def plot(self, atlevel):
        fig = plt.figure()
        
        X = []
        Y = []
        
        for w in range(0, len(self.cost[atlevel][0])):
            for s in range(0, len(self.cost[atlevel])):
                X.append(s*w)
                Y.append(self.cost[atlevel][s][w]/self.period)#self.cost[atlevel][0][0])
        
        XY= zip(X,Y)
        
        XY.sort()
        
        X = [x for x, y in XY]
        Y = [y for x, y in XY]
        
        plt.plot(X, Y)
#         plt.scatter(X, Y)
        
#         ax = fig.gca(projection='3d')
#     
#         S = range(0, len(self.cost[atlevel]))
#         W = range(0, len(self.cost[atlevel][0]))
#         S, W = np.meshgrid(S, W)
#         
#         Z = np.asarray([[self.cost[atlevel][S[i][j]][W[i][j]]/self.cost[atlevel][0][0] for j in range(0, len(W))] for i in range(0, len(S))])
#         
#         ax.plot_wireframe(S, W, Z)
        
        plt.savefig("ex-exec.pdf", bbox_inches='tight', pad_inches=0.1)
        
        plt.close()
        
    def plot2(self):
        fig = plt.figure(figsize=(8,4))
        
        X = []
        Y = []
        
        for w in range(0, len(self.cost['Cshared'][0])):
            for s in range(0, len(self.cost['Cshared'])):
                X.append(4*s*w)
                Y.append(self.cost['Ciso'][s][w]/1000.0)#/self.period)#self.cost[atlevel][0][0])
        
        XY= zip(X,Y)
        
        XY.sort()
        
        X = [x for x, y in XY]
        Y = [y for x, y in XY]
        
        ax = plt.plot(X, Y, '-', label="$\\mathbf{C_i^7}$", linewidth=3.0)
        
#         .axis["bottom", "top", "right"].set_visible(False)

        X = []
        Y = []
        
        for w in range(0, len(self.cost['Cshared'][0])):
            for s in range(0, len(self.cost['Cshared'])):
                X.append(4*s*w)
                Y.append(self.cost['Cshared'][s][w]/1000.0)#/self.period)#self.cost[atlevel][0][0])
        
        XY= zip(X,Y)
        
        XY.sort()
        
        X = [x for x, y in XY]
        Y = [y for x, y in XY]
        
        plt.plot(X, Y, '--', label="$\\mathbf{C_i^8}$", linewidth=3.0)
        
        
        maxCost = self.cost['Ciso'][0][0]/1000.0
        minCost = self.cost['Ciso'][16][16]/1000.0
        plt.axis([0, 1024, minCost - 0.8, maxCost+2.5])
        
        plt.legend(loc="upper right", ncol = 2)
        
        plt.ylabel("ACET Parameter Value (ms)")
        plt.xlabel("Allocated Cache Space (KB)")
        
        #adjust plot
        plt.subplots_adjust(left=0.15)
        plt.xticks([0, 256, 512, 768, 1024])
        
        plt.axhline(y = maxCost, linestyle = ':', linewidth = 3, color = 'gray')
        plt.axhline(y = minCost, linestyle = ':', linewidth = 3, color = 'gray')
        
        plt.axes([0.95, .15, 0.001, .59], axisbg='w')
        
        plt.ylabel("Reduction")
        
        plt.tick_params(
        axis='x',         
        which='both',    
        bottom='off',   
        top='off',
        left='off',
        labelbottom='off',
#         labelleft = 'off',
#         labelright = 'off',
        direction = 'inout')
        
        
        fmt = '%.0f%%' # Format you want the ticks, e.g. '40%'
        yticks = mtick.FormatStrFormatter(fmt)
        plt.ylim(0, 100)
        plt.gca().yaxis.set_major_formatter(yticks)
        
        plt.gca().yaxis.tick_right()
        plt.gca().invert_yaxis()
        plt.gca().yaxis.set_label_position("right")
        plt.gca().yaxis.set_tick_params(width=1.5, size=9)

        plt.savefig("ex-exec.pdf", bbox_inches='tight', pad_inches=0.1)
        
        plt.close()

class TaskSystemMCX(list):
    def __init__(self, tasks=[]):
        self.extend(tasks)
        self.scheme = default_scheme()
        
    def __str__(self):
        return "\n".join([str(t) for t in self])

    def __repr__(self):
        return "TaskSystem([" + ", ".join([repr(t) for t in self]) + "])"
    
    def only_implicit_deadlines(self):
        return forall(self)(lambda t: t.implicit_deadline())

    def only_constrained_deadlines(self):
        return forall(self)(lambda t: t.constrained_deadline())

    def assign_ids(self):
        for i, t in enumerate(self):
            t.id = i + 1

    def assign_ids_by_period(self):
        for i, t in enumerate(sorted(self, key=lambda t: t.period)):
            t.id = i + 1

    def assign_ids_by_deadline(self):
        for i, t in enumerate(sorted(self, key=lambda t: t.deadline)):
            t.id = i + 1

    def sort_by_period(self):
        self.sort(key=lambda t: t.period)

    def sort_by_deadline(self):
        self.sort(key=lambda t: t.deadline)
        
    def useCiso(self):
        for t in self.level('C'):
            t.cost['C'] = [[t.cost['Ciso'][s][w] for w in range(0, self.wg*self.scheme.mW+1)] for s in range(0, self.scheme.mS+1)]

    def useCshared(self):
        for t in self.level('C'):
            t.cost['C'] = [[t.cost['Cshared'][s][w] for w in range(0, self.wg*self.scheme.mW+1)] for s in range(0, self.scheme.mS+1)]
                
    def utilization(self, atlevel, approximate, inflated, c_taskset, minAperiods, mx_wss, heaviest=None):
        u = [t.utilization(atlevel, approximate, inflated, minAperiods[t.core], c_taskset, mx_wss['B'][t.core]) for t in self if t.level is 'A' or t.level is 'B']
        u.extend([t.utilization(atlevel, approximate, inflated, 0, c_taskset, mx_wss['C']) for t in self if t.level is 'C'])
        if heaviest is None:
            return sum(u)
        else:
            u.sort(reverse=True)
            return sum(u[:heaviest])
            
#     #U_A, dictionary of A task set utilizations on individual core for each criticality level
#     #Inflate utilizations based on LLC-allocation dependent inflations
    def secondaryInflateCPD(self, U_A, minAperiods):
        tckinfw = ohc.PTCK/(ohc.PTCK-ohc.ohcvals['Max']['TCK'])
        tckinfa = ohc.PTCK/(ohc.PTCK-ohc.ohcvals['Avg']['TCK'])
        
        for c in range(0, self.m):
            if minAperiods[c] > 0 and self.scheme.WO(c) > 0:
                invminp = 1.0/minAperiods[c]
                U_A['B'][c] += ohc.CPMD*self.scheme.B[c].W*self.scheme.B[c].S*invminp*tckinfw
                U_A['C'][c] += ohc.CPMDavg*self.scheme.WO(c)*self.scheme.B[c].S*invminp*tckinfa

        return U_A
    
    #U_A, dictionary of A task set utilizations on individual core for each criticality level
    #Inflate utilizations based on constant secondary inflations
    def secondaryInflateConst(self, U_A, minAperiods):
        tckinfw = ohc.PTCK/(ohc.PTCK-ohc.ohcvals['Max']['TCK'])
        tckinfa = ohc.PTCK/(ohc.PTCK-ohc.ohcvals['Avg']['TCK'])
        
        for c in range(0, self.m):
            if minAperiods[c] > 0:
                invminp = 1.0/minAperiods[c]
                U_A['B'][c] += (ohc.ohcvals['Max']['CXS']+ohc.ohcvals['Max']['SCH'])*invminp*tckinfw
                U_A['C'][c] += (ohc.ohcvals['Avg']['CXS']+ohc.ohcvals['Avg']['SCH'])*invminp*tckinfa
                
        return U_A
            
    def get_sched_app(self):
        sched = {'A' : [], 'B' : [], 'C' : False, 'H' : False}
        IA_B = [0, 0, 0, 0]
        IA_C = [0, 0, 0, 0]
        for c in range(0, self.m):
            sched['A'].append(self.level('A').core(c).utilizationI_app('A') <= 1)
            if self.scheme.WO(c) > 0:
                invminp = 1.0/min([t.period for t in self.level('A').core(c)])
                IA_B[c] = CPMD*self.scheme.B[c]*self.scheme.S[c]*invminp
                IA_C[c] = CPMDavg*self.scheme.WO(c)*self.scheme.S[c]*invminp
            sched['B'].append(self.level('A').core(c).utilizationI_app('B') + self.level('B').core(c).utilizationI_app('B') + IA_B[c] <= 1)
            
        sched['C'] = self.utilizationI_app('C') + sum(IA_C) <= CORES
        eps = 1e-6
        sched['H'] = self.level('A').utilizationI_app('C') + self.level('B').utilizationI_app('C') + sum(IA_C) + 3*self.hI_app() + self.HI_app() <= CORES-eps
        return sched
        
    def get_sched_all(self):
        sched = {'A' : [True for i in range(0, self.m)], 'B' : [True for i in range(0, self.m)], 'C' : True, 'H' : True}
        return sched

#     def utilization_q(self, heaviest=None):
#         u = [t.utilization_q() for t in self]
#         if heaviest is None:
#             return sum(u)
#         else:
#             u.sort(reverse=True)
#             return sum(u[:heaviest])
# 
#     def density(self):
#         return sum([t.density() for t in self])
# 
#     def density_q(self):
#         return sum([t.density_q() for t in self])

    def hyperperiod(self):
        return lcm(*[t.period for t in self])

    def hI(self, approximate, inflated):
        return self.level('C').max_utilization('C', approximate, inflated, self.level('C'))
        
    def HI(self, approximate, inflated):
        return self.level('C').utilization('C', approximate, inflated, self.level('C'), heaviest=self.m-1)
        
    def max_utilization(self, atlevel, approximate, inflated, c_taskset, minAperiods, mx_wss):
        u = [t.utilization(atlevel, approximate, inflated, minAperiods[t.core], c_taskset, mx_wss) for t in self if t.level is 'A' or t.level is 'B']
        u.extend([t.utilization(atlevel, approximate, inflated, 0, c_taskset, mx_wss) for t in self if t.level is 'C'])
        return mmax(u)

#     def max_density(self):
#         return mmax([t.density() for t in self])
# 
#     def max_density_q(self):
#         return mmax([t.density_q() for t in self])

    def max_cost(self):
        return mmax([t.cost() for t in self])

    def max_period(self):
        return mmax([t.period for t in self])

    def min_deadline(self):
        return min([t.deadline for t in self])

    def max_wss(self):
        "Assumes t.wss has been initialized for each task."
        return mmax([t.wss for t in self])

    def __copy__(self):
        ts = TaskSystem((copy.deepcopy(t) for t in self))
        return ts

    # for backwards compatibility
    copy = __copy__

    def without(self, excluded_tasks):
        "Iterate over contained tasks, skipping over excluded"
        if isinstance(excluded_tasks, SporadicTask):
            # special case: single argument is a task => singleton set
            return (task for task in self if task != excluded_tasks)
        else:
            # general case: caller provided set of tasks to be excluded
            return (task for task in self if not task in excluded_tasks)

    def with_higher_priority_than(self, lower):
        """Iterate over contained tasks, skipping over tasks with priority
        lower than lower.id (i.e., over tasks with larger indices.).
        """
         # assumption: lower id == higher priority
        return (task for task in self if task.id < lower.id)

    def with_lower_priority_than(self, upper):
        """Iterate over contained tasks, skipping over tasks with priority
        higher than upper.id (i.e., over tasks with smaller indices.).
        """
         # assumption: lower id == higher priority
        return (task for task in self if task.id > upper.id)
    
    def level(self, level):
        sts = TaskSystemMCX([t for t in self if t.level is level])
        sts.scheme = self.scheme
        return sts
        
    def core(self, core):
        sts = TaskSystemMCX([t for t in self if t.core is core])
        sts.scheme = self.scheme
        return sts
    
    def setScheme(self, sch):
        self.scheme = scheme(sch.A[:], sch.B[:], sch.C, sch.mW, sch.mS)
        for t in self.level('A'):
            t.ways = sch.A[t.core].W
            t.sets = sch.A[t.core].S
        for t in self.level('B'):
            t.ways = sch.B[t.core].W
            t.sets = sch.B[t.core].S
        for t in self.level('C'):
            t.ways = sch.C.W
            t.sets = sch.C.S
            
    def setLLCAlloc(self, sch): #Set LLC allocations for each Level-L task uniformly
        for t in self.level('A'):
            t.ways = sch.A[0].W
            t.sets = sch.A[0].S
        for t in self.level('B'):
            t.ways = sch.B[0].W
            t.sets = sch.B[0].S
        for t in self.level('C'):
            t.ways = sch.C.W
            t.sets = sch.C.S

    def coreAssign(self, useMC2):
        coreAs = []
        coreUtil = np.zeros(self.m)
        
        if useMC2:
            for t in self.level('A'):
                val, idx = min((val, idx) for (idx, val) in enumerate(coreUtil))
                coreUtil[idx] += t.cost['C'][t.sets][t.ways]/t.period
                t.core = idx
            for t in self.level('B'):
                val, idx = min((val, idx) for (idx, val) in enumerate(coreUtil))
                coreUtil[idx] += t.cost['C'][t.sets][t.ways]/t.period
                t.core = idx
        else:
            for t in self.level('B'):
                val, idx = min((val, idx) for (idx, val) in enumerate(coreUtil))
                coreUtil[idx] += t.cost['B'][t.sets][t.ways]/t.period
                t.core = idx

    def plot(self, level, costlevel, filename):
        
        plt.figure(figsize=(8,4))
        
        LINE_STYLE = ['--', '-.', '-', 'c--', 'k-.']
        for idx, t in enumerate(self.level(level)):
            X = []
            Y = []

            for w in range(0, len(t.cost[costlevel][0])):
                for s in range(0, len(t.cost[costlevel])):
                    X.append(4*s*w)
                    Y.append(t.cost[costlevel][s][w]/t.period)
                    
            XY= zip(X,Y)
        
            XY.sort()
        
            X = [x for x, y in XY]
            Y = [y for x, y in XY]
            
            rc('text', usetex=True)
            
            plt.plot(X, Y, LINE_STYLE[idx], label = "$\\mathbf{\\tau_"+str(idx)+"}$", linewidth = 3)
        
        
        plt.legend(loc="upper right")
        plt.ylabel("Original Task Utilizations")
        plt.xlabel("Allocated Cache Space (KB)")
        
        plt.legend(loc="upper right", ncol = 3)
        plt.xticks([0, 256, 512, 768, 1024])
        plt.xlim([0, 1024])
        
        plt.savefig(filename, bbox_inches='tight', pad_inches=0.1)
        
#         plt.scatter(X, Y)
            
        
#         for t in self.level(level):
#             plt.plot([cost/t.cost[costlevel][t.sets][0] for cost in t.cost[costlevel][t.sets]], linewidth=2.0)
#         plt.ylim([0,1.2])
#         plt.xlim([0,16])
#         plt.ylabel("Normalized PETs")
#         plt.xlabel("Ways")

#         plt.show()
        
        plt.close()
        
