from collections import defaultdict
from csv import DictReader

oheadtypes =   {'CXS' : 'CXS', 
                'ISR' : 'ISR',
                'SCH' : 'SCHED_A',
                'SCL' : 'SCHED_C',
                'RLA' : 'RELEASE_LATENCY_A',
                'RLC' : 'RELEASE_LATENCY_C',
                'RQA' : 'RELEASE',
                'RQC' : 'RELEASE_C',
                'IPI' : 'SCHED_C',
                'TCK' : 'TICK',
                }

oheadcompl =   {'CXS' : 'monotonic', 
                'ISR' : 'linear',
                'SCH' : 'linear',
                'SCL' : 'linear',
                'RLA' : 'monotonic',
                'RLC' : 'monotonic',
                'RQA' : 'linear',
                'RQC' : 'linear',
                'IPI' : 'monotonic',
                'TCK' : 'monotonic',
                }

schedtypes = {'mc2', 'pedf'}

crittypes = {'Max', 'Avg'}

CPMD = 40.0
CPMDavg = 21.0
PTCK = 10000.0

ohcvals = None
ohfuncs = None

class Ohead():
    def __init__(self, tss, data, complexity):
        self.complexity = complexity
        self.values = []
        maxval = -1
        points = 0
        for i in range(0, len(data)-1):
            maxval = max(maxval, data[i])
            self.values.append([])
            self.values[i].append(tss[i])
            self.values[i].append(maxval)
            
    def val(self, tss):
        if self.complexity is "monotonic":
            return self.valmono(tss)
        if self.complexity is "linear":
            return self.vallinear(tss)
        
    def valmono(self, tss):
        if self.values[0][0] >= tss:
            return self.values[0][1]
            
        if self.values[len(self.values)-1][0] < tss:
            return self.values[len(self.values)-1][1]
        for i in range(1, len(self.values)):
            if(self.values[i][0] >= tss):
                p1 = self.values[i-1]
                p2 = self.values[i]
                bary = float((tss-p1[0]))/float(p2[0]-p1[0])
                return p1[1]+bary*(p2[1]-p1[1])
    
    def vallinear(self, tss):
        if self.values[0][0] >= tss:
            return self.values[0][1]
        
        p1 = None
        p2 = None
        if self.values[len(self.values)-1][0] < tss:
            p1 = self.values[len(self.values)/2]
            p2 = self.values[len(self.values)-1]
        else:
            for i in range(1, len(self.values)):
                if(self.values[i][0] >= tss):
                    p1 = self.values[i-1]
                    p2 = self.values[i]
                    break
                    
        bary = float((tss-p1[0]))/float(p2[0]-p1[0])
        return p1[1]+bary*(p2[1]-p1[1])
                
def SetOheads(mcx_ts, scheduler):
    global ohcvals
    
    tss = len(mcx_ts)
    ohcvals = defaultdict()
    for crit in crittypes:
        ohcvals[crit] = defaultdict()
        for ohd in oheadtypes:
            ohcvals[crit][ohd] = ohfuncs[scheduler][crit][ohd].val(tss)
            if scheduler is 'pedf':
                ohcvals[crit][ohd] *= mcx_ts.a_inflate_choice